@@ -16,6 +16,9 @@
 #include "kexec-ppc.h"
 #include "crashdump-powerpc.h"
 
+#define DEVTREE_CRASHKERNEL_BASE "/proc/device-tree/chosen/linux,crashkernel-base"
+#define DEVTREE_CRASHKERNEL_SIZE "/proc/device-tree/chosen/linux,crashkernel-size"
+
 #ifdef CONFIG_PPC64
 static struct crash_elf_info elf_info64 = {
 class: ELFCLASS64,
@@ -41,6 +44,7 @@ lowmem_limit: MAXMEM,
  * A separate program header is created for backup region
  */
 static struct memory_range *crash_memory_range;
+static int crash_nr_memory_ranges;
 
 /* Define a variable to replace the CRASH_MAX_MEMORY_RANGES macro */
 static int crash_max_memory_ranges;
@@ -51,6 +55,29 @@ static int crash_max_memory_ranges;
  */
 mem_rgns_t usablemem_rgns = {0, NULL};
 
+/* Append a segment to crash_memory_range, splitting it into two if
+ * it contains both lowmem and highmem */
+static void add_crash_memory_range(unsigned long long start,
+		unsigned long long end)
+{
+#ifndef CONFIG_PPC64
+	if (start < elf_info32.lowmem_limit && end > elf_info32.lowmem_limit) {
+		add_crash_memory_range(start, elf_info32.lowmem_limit);
+		add_crash_memory_range(elf_info32.lowmem_limit, end);
+		return;
+	}
+#endif
+
+	if (crash_nr_memory_ranges < crash_max_memory_ranges) {
+		crash_memory_range[crash_nr_memory_ranges].start = start;
+		crash_memory_range[crash_nr_memory_ranges].end = end;
+		crash_memory_range[crash_nr_memory_ranges].type = RANGE_RAM;
+	}
+
+	crash_nr_memory_ranges++;
+}
+
+
 /* Reads the appropriate file and retrieves the SYSTEM RAM regions for whom to
  * create Elf headers. Keeping it separate from get_memory_ranges() as
  * requirements are different in the case of normal kexec and crashdumps.
@@ -65,13 +92,12 @@ mem_rgns_t usablemem_rgns = {0, NULL};
 static int get_crash_memory_ranges(struct memory_range **range, int *ranges)
 {
 
-	int memory_ranges = 0;
 	char device_tree[256] = "/proc/device-tree/";
 	char fname[256];
 	DIR *dir, *dmem;
 	int fd;
 	struct dirent *dentry, *mentry;
-	int i, n, crash_rng_len = 0;
+	int n, crash_rng_len = 0;
 	unsigned long long start, end, cstart, cend;
 
 	crash_max_memory_ranges = max_memory_ranges + 6;
@@ -83,13 +109,11 @@ static int get_crash_memory_ranges(struct memory_range **range, int *ranges)
 		return -1;
 	}
 	memset(crash_memory_range, 0, crash_rng_len);
+	crash_nr_memory_ranges = 0;
 
 #ifndef CONFIG_BOOKE
 	/* create a separate program header for the backup region */
-	crash_memory_range[0].start = BACKUP_SRC_START;
-	crash_memory_range[0].end = BACKUP_SRC_END + 1;
-	crash_memory_range[0].type = RANGE_RAM;
-	memory_ranges++;
+	add_crash_memory_range(BACKUP_SRC_START, BACKUP_SRC_END + 1);
 #endif
 
 	dir = opendir(device_tree);
@@ -128,70 +152,37 @@ static int get_crash_memory_ranges(struct memory_range **range, int *ranges)
 				closedir(dir);
 				goto err;
 			}
-			if (memory_ranges >= (max_memory_ranges + 1)) {
-				/* No space to insert another element. */
-				fprintf(stderr,
-					"Error: Number of crash memory ranges"
-					" excedeed the max limit\n");
-				goto err;
-			}
 #ifndef CONFIG_BOOKE
 			if (start == 0 && end >= (BACKUP_SRC_END + 1))
 				start = BACKUP_SRC_END + 1;
 #endif
 
-			cstart = crash_base;
-			cend = crash_base + crash_size;
 			/*
 			 * Exclude the region that lies within crashkernel.
 			 * If memory limit is set then exclude memory region
 			 * above it.
 			 */
+
 			if (memory_limit) {
 				if (start >= memory_limit)
 					continue;
 				if (end > memory_limit)
 					end = memory_limit;
 			}
-			if (cstart < end && cend > start) {
-				if (start < cstart && end > cend) {
-					crash_memory_range[memory_ranges].start
-						= start;
-					crash_memory_range[memory_ranges].end
-						= cstart;
-					crash_memory_range[memory_ranges].type
-						= RANGE_RAM;
-					memory_ranges++;
-					crash_memory_range[memory_ranges].start
-						= cend;
-					crash_memory_range[memory_ranges].end
-						= end;
-					crash_memory_range[memory_ranges].type
-						= RANGE_RAM;
-					memory_ranges++;
-				} else if (start < cstart) {
-					crash_memory_range[memory_ranges].start
-						= start;
-					crash_memory_range[memory_ranges].end
-						= cstart;
-					crash_memory_range[memory_ranges].type
-						= RANGE_RAM;
-					memory_ranges++;
-				} else if (end > cend) {
-					crash_memory_range[memory_ranges].start
-						= cend;
-					crash_memory_range[memory_ranges].end
-						= end;
-					crash_memory_range[memory_ranges].type
-						= RANGE_RAM;
-					memory_ranges++;
-				}
-			} else {
-				crash_memory_range[memory_ranges].start = start;
-				crash_memory_range[memory_ranges].end  = end;
-				crash_memory_range[memory_ranges].type
-					= RANGE_RAM;
-				memory_ranges++;
+
+			/*
+			 * Exclure region used by crash kernel
+			 */
+			cstart = crash_base;
+			cend = crash_base + crash_size;
+
+			if (cstart >= end || cend <= start)
+				add_crash_memory_range(start, end);
+			else {
+				if (start < cstart)
+					add_crash_memory_range(start, cstart);
+				if (cend < end)
+					add_crash_memory_range(cend, end);
 			}
 		}
 		closedir(dmem);
@@ -210,12 +201,18 @@ static int get_crash_memory_ranges(struct memory_range **range, int *ranges)
 			cstart = crash_base;
 		if (cend > crash_base + crash_size)
 			cend = crash_base + crash_size;
-		crash_memory_range[memory_ranges].start = cstart;
-		crash_memory_range[memory_ranges++].end = cend;
+		add_crash_memory_range(cstart, cend);
+	}
+
+	if (crash_nr_memory_ranges >= crash_max_memory_ranges) {
+		fprintf(stderr,
+			"Error: Number of crash memory ranges"
+			" excedeed the max limit\n");
+		goto err;
 	}
 
 	*range = crash_memory_range;
-	*ranges = memory_ranges;
+	*ranges = crash_nr_memory_ranges;
 
 	int j;
 	dbgprintf("CRASH MEMORY RANGES\n");
@@ -403,11 +400,28 @@ void add_usable_mem_rgns(unsigned long long base, unsigned long long size)
 		usablemem_rgns.size, base, size);
 }
 
+int get_crash_kernel_load_range(uint64_t *start, uint64_t *end)
+{
+	unsigned long long value;
+
+	if (!get_devtree_value(DEVTREE_CRASHKERNEL_BASE, &value))
+		*start = value;
+	else
+		return -1;
+
+	if (!get_devtree_value(DEVTREE_CRASHKERNEL_SIZE, &value))
+		*end = *start + value - 1;
+	else
+		return -1;
+
+	return 0;
+}
+
 int is_crashkernel_mem_reserved(void)
 {
 	int fd;
 
-	fd = open("/proc/device-tree/chosen/linux,crashkernel-base", O_RDONLY);
+	fd = open(DEVTREE_CRASHKERNEL_BASE, O_RDONLY);
 	if (fd < 0)
 		return 0;
 	close(fd);
